@@ -25,10 +25,11 @@
 
 
 __all__ = ('Connector', 'RPCSession', 'MessageSession', 'Server', 'ExcessiveSessionCostError',
-           'BatchError')
+           'BatchError', 'Concurrency')
 
 
 import asyncio
+from contextlib import suppress
 import logging
 from math import ceil
 import time
@@ -55,8 +56,11 @@ class Connector(object):
         self.host = host
         self.port = port
         self.proxy = proxy
+        self.protocol = None
         self.loop = kwargs.get('loop', asyncio.get_event_loop())
         self.kwargs = kwargs
+        # By default, do not limit outgoing connections
+        self.cost_hard_limit = 0
 
     async def create_connection(self):
         '''Initiate a connection.'''
@@ -66,8 +70,6 @@ class Connector(object):
 
     async def __aenter__(self):
         _transport, self.protocol = await self.create_connection()
-        # By default, do not limit outgoing connections
-        self.ru_hard_limit = 0
         return self.protocol
 
     async def __aexit__(self, exc_type, exc_value, traceback):
@@ -86,15 +88,11 @@ class Concurrency(object):
         self._sem_value = self._target
 
     async def _retarget_semaphore(self):
-        while self._sem_value != self._target:
-            if self._target <= 0:
-                raise ExcessiveSessionCostError
-            if self._sem_value > self._target:
-                await self._semaphore.acquire()
-                self._sem_value -= 1
-            else:
-                self._semaphore.release()
-                self._sem_value += 1
+        if self._target <= 0:
+            raise ExcessiveSessionCostError
+        while self._sem_value < self._target:
+            self._sem_value += 1
+            self._semaphore.release()
 
     @property
     def max_concurrent(self):
@@ -104,11 +102,14 @@ class Concurrency(object):
         self._target = int(target)
 
     async def __aenter__(self):
-        await self._retarget_semaphore()
         await self._semaphore.acquire()
+        await self._retarget_semaphore()
 
     async def __aexit__(self, exc_type, exc_value, traceback):
-        self._semaphore.release()
+        if self._sem_value > self._target:
+            self._sem_value -= 1
+        else:
+            self._semaphore.release()
 
 
 class SessionBase(asyncio.Protocol):
@@ -125,19 +126,24 @@ class SessionBase(asyncio.Protocol):
     on entry to the block, and closed on exit from the block.
     '''
 
-    max_errors = 10
     # Multiply this by bandwidth bytes used to get resource usage cost
     bw_cost_per_byte = 1 / 100000
     # If cost is over this requests begin to get delayed and concurrency is reduced
-    cost_soft_limit = 10000
+    cost_soft_limit = 2000
     # If cost is over this the session is closed
-    cost_hard_limit = 50000
+    cost_hard_limit = 10000
     # Resource usage is reduced by this every second
     cost_decay_per_sec = cost_hard_limit / 3600
     # Request delay ranges from 0 to this between cost_soft_limit and cost_hard_limit
-    cost_sleep = 5.0
+    cost_sleep = 2.0
+    # Base cost of an error.  Errors that took resources to discover incur additional costs
+    error_base_cost = 100.0
     # Initial number of requests that can be concurrently processed
-    initial_concurrent = 6
+    initial_concurrent = 20
+    # Send a "server busy" error if processing a request takes longer than this seconds
+    processing_timeout = 30.0
+    # Force-close a connection if its socket send buffer stays full this long
+    max_send_delay = 20.0
 
     def __init__(self, *, framer=None, loop=None):
         self.framer = framer or self.default_framer()
@@ -153,9 +159,8 @@ class SessionBase(asyncio.Protocol):
         # Cleared when the send socket is full
         self._can_send = self.event()
         self._can_send.set()
-        self._task_group = TaskGroup()
-        # Force-close a connection if a send doesn't succeed in this time
-        self.max_send_delay = 60
+        self._task = None
+        self._group = TaskGroup()
         # Statistics.  The RPC object also keeps its own statistics.
         self.start_time = time.time()
         self.errors = 0
@@ -170,9 +175,12 @@ class SessionBase(asyncio.Protocol):
         self._cost_last = 0.0
         self._cost_time = self.start_time
         self._cost_fraction = 0.0
-        # Concurrency control
-        self._concurrency = Concurrency(self.initial_concurrent)
-        self._cost_check_delta = max(self.cost_soft_limit // 4, 100)
+        # Concurrency control for incoming request handling
+        self._incoming_concurrency = Concurrency(self.initial_concurrent)
+
+    async def _process_messages(self):
+        async with self._group:
+            await self._group.spawn(self._receive_messages)
 
     def _receive_messages(self):
         raise NotImplementedError
@@ -198,12 +206,11 @@ class SessionBase(asyncio.Protocol):
             if self.verbosity >= 4:
                 self.logger.debug(f'Sending framed message {framed_message}')
             self.transport.write(framed_message)
+        return self.last_send
 
-    def _bump_errors(self):
+    def _bump_errors(self, exception=None):
         self.errors += 1
-        if self.errors >= self.max_errors:
-            # Don't await self.close() because that is self-cancelling
-            self._close()
+        self.bump_cost(self.error_base_cost + getattr(exception, 'cost', 0.0))
 
     def _close(self):
         if self.transport:
@@ -244,7 +251,7 @@ class SessionBase(asyncio.Protocol):
             self._proxy_address = peer_address
         else:
             self._address = peer_address
-        self._task = spawn_sync(self._receive_messages(), loop=self.loop)
+        self._task = spawn_sync(self._process_messages(), loop=self.loop)
 
     def connection_lost(self, exc):
         '''Called by asyncio when the connection closes.
@@ -259,10 +266,14 @@ class SessionBase(asyncio.Protocol):
         self._can_send.set()
 
     # External API
+    def is_send_buffer_full(self):
+        '''Return True if the send socket buffer is full.'''
+        return not self._can_send.is_set()
+
     def bump_cost(self, delta):
         # Delta can be positive or negative
         self.cost = max(0, self.cost + delta)
-        if abs(self.cost - self._cost_last) > self._cost_check_delta:
+        if abs(self.cost - self._cost_last) > 100:
             self.recalc_concurrency()
 
     def recalc_concurrency(self):
@@ -277,7 +288,7 @@ class SessionBase(asyncio.Protocol):
         self._cost_last = self.cost
 
         # Setting cost_hard_limit <= 0 means to not limit concurrency
-        value = self._concurrency.max_concurrent
+        value = self._incoming_concurrency.max_concurrent
         cost_soft_range = self.cost_hard_limit - self.cost_soft_limit
         if cost_soft_range <= 0:
             return
@@ -286,9 +297,14 @@ class SessionBase(asyncio.Protocol):
         self._cost_fraction = max(0.0, (cost - self.cost_soft_limit) / cost_soft_range)
 
         target = max(0, ceil((1.0 - self._cost_fraction) * self.initial_concurrent))
-        if target != value:
+        if abs(target - value) > 1:
             self.logger.info(f'changing task concurrency from {value} to {target}')
-        self._concurrency.set_target(target)
+        self._incoming_concurrency.set_target(target)
+
+    def unanswered_request_count(self):
+        '''The number of requests received but not yet answered.'''
+        # Max with zero in case the message processing task hasn't yet spawned
+        return max(0, len(self._group._pending) - 1)
 
     def extra_cost(self):
         '''A dynamic value added to this session's cost when deciding how much to throttle
@@ -335,10 +351,13 @@ class SessionBase(asyncio.Protocol):
     async def close(self, *, force_after=30):
         '''Close the connection and return when closed.'''
         self._close()
-        async with ignore_after(force_after):
+        if self._task:
+            async with ignore_after(force_after):
+                await self.closed_event.wait()
+            self.abort()
             await self.closed_event.wait()
-        self.abort()
-        await self.closed_event.wait()
+            with suppress(CancelledError):
+                await self._task
 
 
 class MessageSession(SessionBase):
@@ -358,6 +377,7 @@ class MessageSession(SessionBase):
                     f'bad network magic: got {magic} expected {expected}, '
                     f'disconnecting'
                 )
+                self._bump_errors(e)
                 self._close()
             except OversizedPayloadError as e:
                 command, payload_len = e.args
@@ -365,6 +385,7 @@ class MessageSession(SessionBase):
                     f'oversized payload of {payload_len:,d} bytes to command '
                     f'{command}, disconnecting'
                 )
+                self._bump_errors(e)
                 self._close()
             except BadChecksumError as e:
                 payload_checksum, claimed_checksum = e.args
@@ -372,21 +393,26 @@ class MessageSession(SessionBase):
                     f'checksum mismatch: actual {payload_checksum.hex()} '
                     f'vs claimed {claimed_checksum.hex()}'
                 )
-                self._bump_errors()
+                self._bump_errors(e)
             else:
                 self.last_recv = time.time()
                 self.recv_count += 1
-                await self._throttled_message(message)
+                await self._group.spawn(self._throttled_message(message))
 
     async def _throttled_message(self, message):
         '''Process a single request, respecting the concurrency limit.'''
         try:
-            async with self._concurrency:
-                if self._cost_fraction:
-                    await sleep(self._cost_fraction * self.cost_sleep)
-                await self.handle_message(message)
+            timeout = self.processing_timeout
+            async with timeout_after(timeout):
+                async with self._incoming_concurrency:
+                    if self._cost_fraction:
+                        await sleep(self._cost_fraction * self.cost_sleep)
+                    await self.handle_message(message)
         except ProtocolError as e:
             self.logger.error(f'{e}')
+            self._bump_errors(e)
+        except TaskTimeout:
+            self.logger.info(f'incoming request timed out after {timeout} secs')
             self._bump_errors()
         except ExcessiveSessionCostError:
             self._close()
@@ -472,15 +498,9 @@ class BatchRequest(object):
         if exc_type is None:
             self.batch = Batch(self._requests)
             message, event = self._session.connection.send_batch(self.batch)
-            await self._session._send_message(message)
-            await event.wait()
-            result = event.result
-            # Can happen with cancel_pending_requests
-            if isinstance(result, CancelledError):
-                raise result
-            self.results = result
+            self.results = await self._session._send_concurrent(message, event, len(self.batch))
             if self._raise_errors:
-                if any(isinstance(item, Exception) for item in result):
+                if any(isinstance(item, Exception) for item in self.results):
                     raise BatchError(self)
 
 
@@ -488,9 +508,32 @@ class RPCSession(SessionBase):
     '''Base class for protocols where a message can lead to a response,
     for example JSON RPC.'''
 
+    # Adjust outgoing request concurrency to target a round trip response time of
+    # this many seconds, recalibrating every recalibrate_count requests
+    target_response_time = 3.0
+    recalibrate_count = 30
+    # Raise a TaskTimeout if getting a response takes longer than this
+    sent_request_timeout = 30.0
+    log_me = False
+
     def __init__(self, *, framer=None, loop=None, connection=None):
         super().__init__(framer=framer, loop=loop)
         self.connection = connection or self.default_connection()
+        # Concurrency control for outgoing request sending
+        self._outgoing_concurrency = Concurrency(50)
+        self._req_times = []
+
+    def _recalc_concurrency(self):
+        req_times = self._req_times
+        avg = sum(req_times) / len(req_times)
+        req_times.clear()
+        current = self._outgoing_concurrency.max_concurrent
+        cap = min(current + max(3, current * 0.1), 250)
+        floor = max(1, min(current * 0.8, current - 1))
+        target = int(0.5 + max(floor, min(cap, current * self.target_response_time / avg)))
+        if target != current:
+            self.logger.info(f'changing outgoing request concurrency to {target} from {current}')
+            self._outgoing_concurrency.set_target(target)
 
     async def _receive_messages(self):
         while not self.is_closing():
@@ -502,48 +545,75 @@ class RPCSession(SessionBase):
 
             self.last_recv = time.time()
             self.recv_count += 1
+            if self.log_me:
+                self.logger.info(f'processing {message}')
 
             try:
                 requests = self.connection.receive_message(message)
             except ProtocolError as e:
-                self.logger.debug(f'{e}')
+                self.logger.debug(str(e))
+                if e.code == JSONRPC.PARSE_ERROR:
+                    e.cost = self.error_base_cost * 10
+                self._bump_errors(e)
                 if e.error_message:
                     await self._send_message(e.error_message)
-                if e.code == JSONRPC.PARSE_ERROR:
-                    self.max_errors = 0
-                self._bump_errors()
             else:
                 for request in requests:
-                    await self._throttled_request(request)
+                    await self._group.spawn(self._throttled_request(request))
 
     async def _throttled_request(self, request):
         '''Process a single request, respecting the concurrency limit.'''
         try:
-            async with self._concurrency:
-                if self._cost_fraction:
-                    await sleep(self._cost_fraction * self.cost_sleep)
-                result = await self.handle_request(request)
+            timeout = self.processing_timeout
+            async with timeout_after(timeout):
+                async with self._incoming_concurrency:
+                    if self._cost_fraction:
+                        await sleep(self._cost_fraction * self.cost_sleep)
+                    result = await self.handle_request(request)
         except (ProtocolError, RPCError) as e:
             result = e
+        except TaskTimeout:
+            self.logger.info(f'incoming request {request} timed out after {timeout} secs')
+            result = RPCError(JSONRPC.SERVER_BUSY, 'server busy - request timed out')
         except ExcessiveSessionCostError:
             result = FinalRPCError(JSONRPC.EXCESSIVE_RESOURCE_USAGE, 'excessive resource usage')
         except CancelledError:
             raise
         except Exception:
             self.logger.exception(f'exception handling {request}')
-            result = RPCError(JSONRPC.INTERNAL_ERROR,
-                              'internal server error')
+            result = RPCError(JSONRPC.INTERNAL_ERROR, 'internal server error')
 
         if isinstance(request, Request):
             message = request.send_result(result)
             if message:
                 await self._send_message(message)
         if isinstance(result, Exception):
-            self._bump_errors()
+            self._bump_errors(result)
             if isinstance(result, FinalRPCError):
                 # Don't await self.close() because that is self-cancelling
                 self._close()
 
+    async def _send_concurrent(self, message, event, request_count):
+        async with self._outgoing_concurrency:
+            send_time = await self._send_message(message)
+            try:
+                async with timeout_after(self.sent_request_timeout):
+                    await event.wait()
+            finally:
+                time_taken = time.time() - send_time
+                if request_count == 1:
+                    self._req_times.append(time_taken)
+                else:
+                    self._req_times.extend([time_taken / request_count] * request_count)
+                if len(self._req_times) >= self.recalibrate_count:
+                    self._recalc_concurrency()
+
+        result = event.result
+        # For batches, CancelledError can happen with cancel_pending_requests
+        if isinstance(result, Exception):
+            raise result
+        return result
+
     def connection_lost(self, exc):
         # Cancel pending requests and message processing
         self.connection.cancel_pending_requests()
@@ -564,12 +634,7 @@ class RPCSession(SessionBase):
     async def send_request(self, method, args=()):
         '''Send an RPC request over the network.'''
         message, event = self.connection.send_request(Request(method, args))
-        await self._send_message(message)
-        await event.wait()
-        result = event.result
-        if isinstance(result, Exception):
-            raise result
-        return result
+        return await self._send_concurrent(message, event, 1)
 
     async def send_notification(self, method, args=()):
         '''Send an RPC notification over the network.'''
