@@ -23,18 +23,17 @@
 # OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 # WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 
-__all__ = ('normalize_corofunc', )
+__all__ = ('instantiate_coroutine', )
 
 
+import asyncio
 from collections import namedtuple
 from functools import partial
 import inspect
 
-from aiorpcx import Semaphore
 
-
-def normalize_corofunc(corofunc, args):
-    if inspect.iscoroutine(corofunc):
+def instantiate_coroutine(corofunc, args):
+    if asyncio.iscoroutine(corofunc):
         if args != ():
             raise ValueError('args cannot be passed with a coroutine')
         return corofunc
@@ -86,34 +85,6 @@ def signature_info(func):
     return SignatureInfo(min_args, max_args, required_names, other_names)
 
 
-def _require_non_negative(value):
-    if not isinstance(value, int) or value < 0:
-        raise RuntimeError('concurrency must be a natural number')
-
-
-class Concurrency(object):
-
-    def __init__(self, max_concurrent):
-        _require_non_negative(max_concurrent)
-        self._max_concurrent = max_concurrent
-        self.semaphore = Semaphore(max_concurrent)
-
-    @property
-    def max_concurrent(self):
-        return self._max_concurrent
-
-    async def set_max_concurrent(self, value):
-        _require_non_negative(value)
-        diff = value - self._max_concurrent
-        self._max_concurrent = value
-        if diff >= 0:
-            for _ in range(diff):
-                self.semaphore.release()
-        else:
-            for _ in range(-diff):
-                await self.semaphore.acquire()
-
-
 def check_task(logger, task):
     if not task.cancelled():
         try:
